from lwan_coro cimport scheduler_t, scheduler_init, scheduler_coro_add, scheduler_coro_add_runtime, scheduler_run, async_scheduler_run, scheduler_destroy, async_scheduler_destroy, coro_yield_value, coro_yield, coro_t
from libc.stdio cimport printf
include "minimal_lib.pxi"

cdef cypclass A nolock activable:
  int a

  void foo(self):
    printf("Foo call with object number %d\n", self.a)

  void bar(self):
    printf("Bar call with object number %d\n", self.a)

  int baz(self):
    printf("Baz call with object number %d\n", self.a)
    # Compute square
    return self.multiply(self.a)

  int multiply(self, int arg):
    return self.a*arg

  int modulo_add(self, int arg, int modulo = 10):
    return (self.a + arg) % modulo

cdef cypclass B(A) activable:
  int add(self, int arg):
    return self.a + arg

cdef ActhonResultInterface take_activated_object(A.Activated actor) nogil:
  return actor.multiply(NULL, 4)

cdef void set_A_attrs(A obj):
  obj._active_result_class = WaitResult.construct
  obj._active_queue_class = BasicQueue()


cdef int process_queue(coro_t *coroutine, void *arg) nogil:
  queue = <BasicQueue> arg
  cdef bint remaining_messages = True
  cdef bint res = False
  while remaining_messages:
    with wlocked queue:
      res = queue.activate()
      if not res:
        remaining_messages = not queue.is_empty()
    printf("Worker, yielding\n")
    coro_yield(coroutine, coro_yield_value.MAY_RESUME)
  del queue
  coro_yield(coroutine, coro_yield_value.FINISHED)

cpdef void main():
  cdef scheduler_t scheduler
  cdef A.Activated actor
  cdef ActivityCounterSync sync = NULL

  scheduler_init(&scheduler, num_workers=2)

  #sync = ActivityCounterSync(sync)

  for i in range(4):
    prev_sync = sync
    sync = ActivityCounterSync(prev_sync)
    obj = A()
    obj.a = i
    set_A_attrs(obj)
    scheduler_coro_add(&scheduler, process_queue, <void*>obj._active_queue_class)
    # Now populate the queue with some messages
    actor = obj.__activate__()
    actor.foo(sync)
    actor.bar(prev_sync)
    # Useful ?
    #Cy_INCREF(prev_sync)
    # So bar will be fired before foo

  res = actor.baz(NULL)
  scheduler_run(&scheduler)
  scheduler_destroy(&scheduler)
  with rlocked res:
    print(<int> res)